from datetime import datetime
from typing import Any, Dict, List, Literal, Optional

from .base import BaseAPIModel


class SillyTeam:
    """Wrapper class for current rewards in the /sillymeter
    
    Attributes
    ----------
    reward : str
        the reward this Silly Team will give

    description : str
        the description of the reward this Silly Team will give

    points : int
        the amount of points this Silly Team earned in the cycle, 
        set to `None` if the Silly Meter state is not set to `Reward`
    """

    __slots__ = ['reward', 'description', 'points']

    def __init__(self, reward: str, description: str, points: Optional[int]) -> None:
        self.reward = reward
        self.description = description
        self.points = points

    def __str__(self) -> str:
        return self.reward

    def __repr__(self) -> str:
        return self.__str__()


class SillyMeter(BaseAPIModel):
    """Wrapper class for /sillymeter response

    A tuple-like class containing `SillyTeam` objects, sorted by points

    Attributes
    ----------
    state : Literal['Active', 'Reward', 'Inactive']
        the current state of the Silly Meter

    hp : int
        the current HP of the Silly Meter (0 - 5,000,000)

    winner : Optional[str]
        the winning Silly Team whose reward is currently active, otherwise `None`

    next_update_timestamp : datetime
        when the Silly Meter will next update itself

    as_of : datetime
        when the server generated the Silly Meter data
    """

    __slots__ = ['state', 'hp', 'winner', 'next_update_timestamp', 'as_of']

    def __init__(self, **payload: Dict[str, Any]) -> None:
        self.state: Literal['Active', 'Reward', 'Inactive'] = payload.get('state')
        self.hp: int = payload.get('hp')

        self.winner: Optional[str] = payload.get('winner')
        self.next_update_timestamp: datetime = datetime.fromtimestamp(payload.get('nextUpdateTimestamp'))
        self.as_of: datetime = datetime.fromtimestamp(payload.get('asOf'))

        rewards: List[str] = payload.get('rewards')
        reward_descriptions: List[str] = payload.get('rewardDescriptions')
        reward_points: List[Optional[int]] = payload.get('rewardPoints')
        iterable = map(lambda args: SillyTeam(*args), zip(rewards, reward_descriptions, reward_points))

        if self.state == 'Reward':
            iterable = tuple(sorted(iterable, key=lambda silly_team: silly_team.points))
        else:
            iterable = tuple(iterable)

        super().__init__(iterable)



from typing import List

from .base import BaseAPIModel


__all__ = ['Doodle', 'Doodles']


class Doodle:
    """Wrapper class for doodle data

    Attributes
    ----------
    district : str
        the district the doodle is in

    playground : str
        the playground the doodle is in

    dna : str
        the doodle's DNA string

    rendition : str
        a link to a 256x256 png of the doodle's rendition

    traits : List[str]
        the list of the doodle's traits

    cost : int
        how much the doodle costs to purchase
    """

    __slots__ = ['district', 'playground', 'dna', 'rendition', 'traits', 'cost']

    def __init__(self, district, playground, *, dna, traits, cost) -> None:
        self.district: str = district
        self.playground: str = playground
        self.dna: str = dna
        self.rendition: str = f'https://rendition.toontownrewritten.com/render/{dna}/doodle/256x256.png'
        self.traits: List[str] = traits
        self.cost: int = cost


class Doodles(BaseAPIModel):
    """Wrapper class for the /doodles response
    
    A tuple-like class containing `Doodle` objects, sorted by district and playground
    """
    def __init__(self, **payload) -> None:
        iterable: tuple[Doodle] = tuple(
            sorted([
                Doodle(district, playground, **doodle)
                for district, playgrounds in payload.items()
                for playground, doodles in playgrounds.items()
                for doodle in doodles
            ], key = lambda doodle: (doodle.district, doodle.playground))
        )

        super().__init__(iterable)



import asyncio

import toontown
import toontown.models


async def main():
    async with toontown.AsyncToontownClient() as client:
        doodles = await client.doodles()

        for doodle in doodles:
            print(doodle.district, doodle.playground, doodle.dna, doodle.rendition, doodle.traits, doodle.cost)


asyncio.run(main())



import re
from datetime import datetime
from typing import Optional

from .base import BaseAPIModel


__all__ = ['News', 'NewsList']


DATE_FMT = '%B %-d, %Y at %-I:%M %p'

cleaner = re.compile(r'(<.*?>|\r)')
encode_cleaner = re.compile(r'&nbsp;')
newline_cleaner = re.compile(r'\n{2,}')


def clean(string: Optional[str]):
    if string is None:
        return string

    string = cleaner.sub('', string)
    string = encode_cleaner.sub(' ', string)
    string = newline_cleaner.sub('\n\n', string)

    return string


class News:
    """Wrapper class for the news data
    
    Attributes
    ----------
    post_id : int
        the ID of the post

    title : str
        the title of the post

    author : str
        who wrote the post

    body_raw : str
        the raw body of the post containing HTML elements

    body : str
        a readable version of the body, with HTML elements stripped out

    date : datetime
        when the post was written

    image : str
        a link to the image of the post
    """

    __slots__ = ['post_id', 'title', 'author', 'body_raw', 'body', 'date', 'image']

    def __init__(self, **props) -> None:
        self.post_id: int = props.get('postId')
        self.title: str = props.get('title')
        self.author: str = props.get('author')
        self.date: datetime = datetime.strptime(props.get('date'), DATE_FMT)
        self.image: str = props.get('image')

        body_raw = props.get('body', None)
        self.body_raw: Optional[str] = body_raw
        self.body: Optional[str] = clean(body_raw)


class NewsList(BaseAPIModel):
    """Wrapper class for the /news response
    
    A tuple-like class containing `News` objects
    """
    def __init__(self, **payload) -> None:
        iterable = tuple(payload.get('iterable'))
        super().__init__(iterable)



from typing import Any, Iterator, Optional, Tuple

from ..exceptions import FailedResponse


__all__ = ['BaseAPIModel']


class BaseAPIModel:
    def __new__(cls, *args, **payload):
        instance = super().__new__(cls)

        if payload.get('error', None):
            """Sometimes the server will send an error field when there is no cached response"""
            raise FailedResponse(payload['error'])
            
        return instance

    def __init__(self, iterable: Tuple[Any]) -> None:
        self.__iterable = iterable

    def count(self, value: Any) -> int:
        return self.__iterable.count(value)
        
    def index(self, value: Any, start: Optional[int] = None, stop: Optional[int] = None) -> int:
        return self.__iterable.index(value, start, stop)

    def __getitem__(self, index: int) -> Any:
        return self.__iterable.__getitem__(index)

    def __len__(self) -> int:
        return self.__iterable.__len__()

    def __iter__(self) -> Iterator[Any]:
        return self.__iterable.__iter__()

    def __next__(self) -> Any:
        return next(self.__iterable)

    def __str__(self) -> str:
        return self.__iterable.__str__()

    def __repr__(self) -> str:
        return self.__iterable.__str__()



from datetime import datetime
from typing import Optional

from .base import BaseAPIModel


__all__ = ['FieldOffice', 'FieldOffices']


zone_lookup = {
    '3100': 'Walrus Way',
    '3200': 'Sleet Street',
    '3300': 'Polar Place',
    '4100': 'Alto Avenue',
    '4200': 'Baritone Boulevard',
    '4300': 'Tenor Terrace',
    '5100': 'Elm Street',
    '5200': 'Maple Street',
    '5300': 'Oak Street',
    '9100': 'Lullaby Lane',
    '9200': 'Pajama Place',
}

department_lookup = {
    's': 'Sellbot'
}


class FieldOffice:
    """Wrapper class for field office data
    
    Attributes
    ----------
    last_updated : datetime
        when the field office's data was last updated

    street : str
        the street that the field office is on

    department : str
        the department of the field office (currently only `Sellbot`)

    difficulty : int
        how many stars the field office is

    annexes : int
        how many annexes are left in the field office

    open : bool
        whether or not the elevator to the field office is open

    expiring : Optional[datetime]
        shows how long left the field office will stand after the last annex is defeated, otherwise `None`
    """

    __slots__ = ['last_updated', 'street', 'department', 'difficulty',
                 'annexes', 'open', 'expiring']
    
    def __init__(self, last_updated, zone, *, department, difficulty, annexes, open, expiring) -> None:
        self.last_updated: datetime = last_updated
        self.street: str = zone_lookup[zone]
        self.department: str = department_lookup[department]
        self.difficulty: int = difficulty + 1
        self.annexes: int = annexes
        self.open: bool = open
        self.expiring: Optional[datetime] = datetime.fromtimestamp(expiring) if expiring else None


class FieldOffices(BaseAPIModel):
    """"Wrapper class for /fieldoffices response

    A tuple-like class containing `FieldOffice` objects, sorted by difficulty

    Attributes
    ----------
    last_updated : datetime
        the time when the field offices data was last updated
    """

    __slots__ = ['last_updated']
    
    def __init__(self, **payload) -> None:
        self.last_updated = last_updated = datetime.fromtimestamp(payload.pop('lastUpdated'))
        iterable = tuple(
            sorted([
                FieldOffice(last_updated, zone, **props)
                for zone, props in payload.pop('fieldOffices').items()
            ],
            key=lambda field_office: field_office.difficulty,
            reverse=True)
        )

        super().__init__(iterable)



class FailedResponse(Exception):
    """An exception that is raised whenever the TTR API returns an error response"""
    
    
class SessionNotConnected(Exception):
    """An exception that is raised whenever an API request is attempted before connecting to the client session"""

    def __init__(self) -> None:
        message = 'Client session is not connected'
        super().__init__(message)
